"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.hasPureNotation = exports.noEffects = exports.isPureFunction = exports.isFirstLetterUpperCase = exports.isFunctionSideEffectFree = exports.getTreeShakingComments = exports.getRootNode = exports.isLocalVariableAWhitelistedModule = exports.getLocalVariable = exports.getChildScopeForNodeIfExists = void 0;
const TREE_SHAKING_COMMENT_ID = "tree-shaking";
const pure_functions_1 = require("../utils/pure-functions");
const getRootNode = (node) => {
    if (node.type === "MemberExpression") {
        return getRootNode(node.object);
    }
    return node;
};
exports.getRootNode = getRootNode;
const getChildScopeForNodeIfExists = (node, currentScope) => currentScope.childScopes.find((scope) => scope.block === node);
exports.getChildScopeForNodeIfExists = getChildScopeForNodeIfExists;
const getLocalVariable = (variableName, scope) => {
    const variableInCurrentScope = scope.variables.find(({ name }) => name === variableName);
    return (variableInCurrentScope ||
        (scope.upper && scope.upper.type !== "global" && getLocalVariable(variableName, scope.upper)));
};
exports.getLocalVariable = getLocalVariable;
const flattenMemberExpressionIfPossible = (node) => {
    switch (node.type) {
        case "MemberExpression":
            if (node.computed || node.property.type !== "Identifier") {
                return null;
            }
            // eslint-disable-next-line no-case-declarations
            const flattenedParent = flattenMemberExpressionIfPossible(node.object);
            return flattenedParent && `${flattenedParent}.${node.property.name}`;
        case "Identifier":
            return node.name;
        default:
            return null;
    }
};
const hasPureNotation = (node, context) => {
    const leadingComments = context.getSourceCode().getCommentsBefore(node);
    if (leadingComments.length) {
        const lastComment = leadingComments[leadingComments.length - 1].value;
        // https://rollupjs.org/configuration-options/#treeshake-annotations
        if (["@__PURE__", "#__PURE__"].includes(lastComment)) {
            return true;
        }
    }
};
exports.hasPureNotation = hasPureNotation;
const isPureFunction = (node, context) => {
    if (hasPureNotation(node, context))
        return true;
    const flattenedExpression = flattenMemberExpressionIfPossible(node);
    if (context.options.length > 0) {
        if (context.options[0].noSideEffectsWhenCalled.find((whiteListedFunction) => whiteListedFunction.function === flattenedExpression)) {
            return true;
        }
    }
    return pure_functions_1.pureFunctions[flattenedExpression];
};
exports.isPureFunction = isPureFunction;
const noEffects = () => { };
exports.noEffects = noEffects;
const parseComment = (comment) => comment.value
    .split(" ")
    .map((token) => token.trim())
    .filter(Boolean);
const getTreeShakingComments = (comments) => {
    const treeShakingComments = comments
        .map(parseComment)
        .filter(([id]) => id === TREE_SHAKING_COMMENT_ID)
        .map((tokens) => tokens.slice(1))
        .reduce((result, tokens) => result.concat(tokens), []);
    return { has: (token) => treeShakingComments.indexOf(token) >= 0 };
};
exports.getTreeShakingComments = getTreeShakingComments;
const isFunctionSideEffectFree = (functionName, moduleName, contextOptions) => {
    if (contextOptions.length === 0) {
        return false;
    }
    for (const whiteListedFunction of contextOptions[0].noSideEffectsWhenCalled) {
        if ((whiteListedFunction.module === moduleName ||
            (whiteListedFunction.module === "#local" && moduleName[0] === ".")) &&
            (whiteListedFunction.functions === "*" ||
                whiteListedFunction.functions.includes(functionName))) {
            return true;
        }
    }
    return false;
};
exports.isFunctionSideEffectFree = isFunctionSideEffectFree;
const isLocalVariableAWhitelistedModule = (variable, property, contextOptions) => {
    if (variable.scope.type === "module" &&
        variable.defs[0].parent &&
        variable.defs[0].parent.source) {
        return isFunctionSideEffectFree(property, variable.defs[0].parent.source.value, contextOptions);
    }
    return false;
};
exports.isLocalVariableAWhitelistedModule = isLocalVariableAWhitelistedModule;
const isFirstLetterUpperCase = (string) => string[0] >= "A" && string[0] <= "Z";
exports.isFirstLetterUpperCase = isFirstLetterUpperCase;
